// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// QGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// QGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with QGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



// Ce programme est libre, vous pouvez le redistribuer et/ou le modifier
// selon les termes de la Licence Publique Gnrale GNU publie par la Free
// Software Foundation (version 2 ou bien toute autre version ultrieure
// choisie par vous).

// Ce programme est distribu car potentiellement utile, mais SANS AUCUNE
// GARANTIE, ni explicite ni implicite, y compris les garanties de
// commercialisation ou d'adaptation dans un but spcifique. Reportez-vous
//  la Licence Publique Gnrale GNU pour plus de dtails.

// Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en
// mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis. 

#ifndef QGLVIEWER_CAMERA_H
#define QGLVIEWER_CAMERA_H

#include "vec.h"
#include "manipulatedFrame.h"
#include "keyFrameInterpolator.h"

class QGLViewer;

namespace qglviewer {
  class Constraint;
  /*!
   * \brief An OpenGL camera description. Includes a frame definition and camera parameters.
   *
   * Camera parameters include the camera type (see Type()) and the field of view.
   * The near and far planes are determined from the scene Radius (updated by the QGLViewer()). 
   */
  class Camera : public QObject
  {
#ifndef DOXYGEN
    friend class ::QGLViewer;
#endif 

    Q_OBJECT        // must include this if you use Qt signals/slots
      
  public:
    Camera();
    /*! Virtual destructor. Empty. */
    virtual ~Camera() {};

    virtual QDomElement domElement(const QString& name, QDomDocument& doc) const;
    virtual void initFromDOMElement(const QDomElement& de);


    /*! The type of camera used by the viewer, set by setType().
     * This type will be use in loadProjectionMatrix().
     * \see setFieldOfView, setSceneRadius, type, setType, loadProjectionMatrix */
    enum Type { PERSPECTIVE, ORTHO };

    /*! @name World to Camera coordinate systems
     *  Methods to go from the world to the camera coordinate systems, and vice-versa.
     */
    //@{
    // Remove all this in version 1.3
  public:
    void getCameraCoordinatesOf(const float src[3], float res[3]) const;
    void getCameraCoordinatesOf(const qglviewer::Vec& src, float res[3]) const;
    /*! Returns the camera frame coordinates of a point \p src defined in world coordinates.
     * worldCoordinatesOf() performs the inverse transformation. */
    qglviewer::Vec cameraCoordinatesOf(const qglviewer::Vec& src) const { return frame.coordinatesOf(src); };
    void getWorldCoordinatesOf(const float src[3], float res[3]) const;
    void getWorldCoordinatesOf(const qglviewer::Vec& src, float res[3]) const;
    /*! Returns the world coordinates of the point whose position is \p src, defined in the camera
     * coordinate system. cameraCoordinatesOf() performs the inverse transformation. */
    qglviewer::Vec worldCoordinatesOf(const qglviewer::Vec& src) const { return frame.inverseCoordinatesOf(src); };
    //@}
    
    /*! @name 2D screen to 3D world coordinate systems
     *  Methods to go from the 2D screen to the 3D world coordinate systems, and vice-versa.
     */
    //@{
    qglviewer::Vec projectedCoordinatesOf(const qglviewer::Vec& src) const;
    void getProjectedCoordinatesOf(const float src[3], float res[3]) const;
    qglviewer::Vec unprojectedCoordinatesOf(const qglviewer::Vec& src) const;
    void getUnprojectedCoordinatesOf(const float src[3], float res[3]) const;
    //@}

    /*! The manipulated frame attached to the camera and which defines its position. */
    qglviewer::ManipulatedFrame frame;


    /*! @name Camera intrinsic parameters
     */
    //@{
public slots:
    /*! Set the displacement mode of the camera, defined from ManipulatedFrame::Mode().
    *
    * See the mouse behavior and customization HTML page for details. */
    void setMode(ManipulatedFrame::Mode m) { frame.setMode(m); };
    /*! Change camera mode (defined by the ManipulatedFrame::Mode() enum), see mode() and setMode(). */
    void toggleMode() { frame.toggleMode(); };

    /*! Defines the vertical field of view of the camera (in degrees). Also used to fit
     * the scene and mouse translations. fieldOfView() returns the current value.
     * The default value is 45 degrees.
     * loadProjectionMatrix() uses this value to define the camera. 
     *
     * The field of view corresponds the one used in \p gluPerspective (see manual).
     * It set the Y (vertical) aperture of the camera. The X (horizontal) angle is
     * infered from the window aspect ratio (see aspectRatio() and horizontalFieldOfView()).
     
     distanceToZeroParallaxPlane() is set to sceneRadius() / tan(fieldOfView()/2) (used only for stereo
     mode, see QGLViewer::setStereo(), setSceneRadius() and setDistanceToZeroParallaxPlane()).
    */
    void setFieldOfView(float f) { fieldOfView_ = f; setDistanceToZeroParallaxPlane(sceneRadius_ / tan(f*M_PI/360.0)); };

    /*! When camera is in FLY mode (see mode()), tunes the speed of the displacement. Current value
     * is given by flySpeed(). Note that is value is also set in setSceneRadius(). */
    void setFlySpeed(float s) { frame.setFlySpeed(s); };
  
    /*! Defines the camera type. Supported types are defined by the Type() enum.
     * loadProjectionMatrix() will create the camera accordingly. Use type() to get the current value.
     \code
     myViewer.camera.setType(qglviewer::Camera::ORTHO);
     \endcode
    */
    void setType(Type type) { type_ = type; };

    /*! This value is used when the camera is \e inside the sceneRadius(), sceneCenter() sphere. The zNear()
      value is then set to zNearCoef() * sceneRadius(). Default value is 0.001, which is good for walkthrough
      applications. In case you need a high dynamic ZBuffer, you can increase this value (~0.1) so that the scene
      uses a wider range of Z values. */
    void setZNearCoef(float coef) { zNearCoef_ = coef; };

    void setSceneRadius(float r);
    void setSceneBoundingBox(const Vec& m, const Vec& M);
    void setSceneCenter(const qglviewer::Vec& sc);
    void setSceneCenter(float x, float y, float z);
    void setSceneCenterFromPixel(int x, int y);

  public:
    /*! Returns the vertical field of view of the camera (in degrees). Value is set using setFieldOfView(). */
    float fieldOfView() const { return fieldOfView_; };

    /*! Returns the type of the camera used by the viewer. Set by setType().
     * Supported types are defined by the Type() enum. Used by loadProjectionMatrix().*/
    Type type() const { return type_; };
  
    /*! Returns the currents camera aspect ratio (screenWidth/screenHeight), determined from the QGLViewer() window size. */
    float aspectRatio() const { return static_cast<float>(screenWidth_)/static_cast<float>(screenHeight_); };

    float distanceToSceneCenter() const;

    /*! Returns the current fly speed (used when mode() is FLY). Set with setFlySpeed().
     * Set to 0.5% of the sceneRadius in setSceneRadius(). */
    float flySpeed() const { return frame.flySpeed(); };
    
    /*! \p true when the camera is being manipulated by the user.
     * Useful for switching to fast draw mode. See QGLViewer::fastDraw(). */
    bool isManipulated() const { return frame.isManipulated(); };

    /*! Returns the value of frame mode, set by setMode() or toggleMode().
     * This mode determines the behavior or the camera when the mouse is moved. */
    ManipulatedFrame::Mode mode() const { return frame.mode(); };

    virtual float zNear() const;
    virtual float zFar()  const;
    /*! Returns the coefficient which is used to set ZNear() when the camera is inside the
      sceneRadius() sphere. Set with setZNearCoef(). */
    float zNearCoef() const { return zNearCoef_; };
        
    /*! Returns the current width (in pixels) of the camera screen. */
    int screenWidth() const { return screenWidth_; };
    /*! Returns the current height (in pixels) of the camera screen. */
    int screenHeight() const { return screenHeight_; };    

    virtual void getProjectionMatrix(GLdouble m[16]) const;
    virtual void getProjectionMatrix(GLfloat m[16]) const;
    virtual void getModelViewMatrix(GLdouble m[16]) const; 
    virtual void getModelViewMatrix(GLfloat m[16]) const;

    qglviewer::Vec pointUnderPixel(const int x, const int y, bool& found);
    //@}

    
    /*! @name Stereo mode
     */
    //@{
public slots:
    /*! Defines the horizontal field of view of the camera (in degrees) by calling
     * setFieldOfView() scaled by 1/aspectRatio() to keep consistency. 
     *
     * loadProjectionMatrixStereo() uses this value to define the camera offset and frustum.*/
    void setHorizontalFieldOfView(float hfov) { setFieldOfView(hfov / aspectRatio()); };
    
    /*! Defines the real inter-ocular distance(in meters). The default value used is 0.062,
     *  which fits most people. 
     *
     * loadProjectionMatrixStereo() uses this value to define the camera offset and frustum*/ 
    void setIODist(float iod) { IODist_ = iod; };

    /*! Defines the distance in real world between the observer eyes and the screen -monitor or projection
     *  wall- in meters. Default value is 0.5m. 
     *
     * loadProjectionMatrixStereo() uses this value to define the camera offset and frustum.*/ 
    void setDistanceToScreen(float dts) { distanceToScreen_ = dts; };

    /*! Defines the distance in virtual world between the viewpoint and the plane where the horizontal
      parallax is null, that's to say where the stereo left and right images are superimposed.  
       
      loadProjectionMatrixStereo() uses this value to define the camera offset and frustum.
      
      \attention This value is changed by QGLViewer::setSceneRadius(), setSceneRadius() and
      setFieldOfView(). When one of these values is modified, distanceToZeroParallaxPlane() is set to
      sceneRadius() / tan(fieldOfView()/2).
    */
    void setDistanceToZeroParallaxPlane (float dtz) { distanceToZeroParallaxPlane_ = dtz; };

    /*! Defines the width of the real screen (in meters). Typically 0.4m (default value) for a monitor
     *  and 10m for the cylindrical wall in Reality Center.
     *
     * loadProjectionMatrixStereo() uses this value to define the camera offset and frustum.*/ 
    void setPhysicalScreenWidth(float sw) { physicalScreenWidth_ = sw; };
  public:
    /*! Returns the horizontal (X camera frame direction) field of view used by the camera (in
      degrees). Value is set using setHorizontalFieldOfView() or setFieldOfView(). The values
      are always linked by : horizontalFieldOfView() = fieldOfView() * Ratio(). */
    float horizontalFieldOfView() const { return aspectRatio() * fieldOfView(); };

    /*! Returns the inter-ocular distance used for stereo. Value set using setIODist() */
    float IODist() const { return IODist_; };

    /*! Returns the physical distance to screen (used for stereo mode only). Value is set using
      setDistanceToScreen() */
    float distanceToScreen() const { return distanceToScreen_; };

    /*! Returns the distance to 0 parallax plane (used for stereo mode only). Value is set using
      setDistanceToZeroParallaxPlane() */
    float distanceToZeroParallaxPlane() const { return distanceToZeroParallaxPlane_; };

    /*! Returns the physical screen width (used for stereo mode only). Value is set using
      setScreenWidth() */
    float physicalScreenWidth() const { return physicalScreenWidth_; };
    //@}

    
    /*! @name Position and orientation of the camera
     */
    //@{
 public slots:
    /*! Same as setPosition(), but with float parameters. */
    void setPosition(float x, float y, float z) { setPosition(qglviewer::Vec(x,y,z)); };
    /*! Set the camera position (the eye), defined in the world coordinate system. The current
      camera position can be obtained from position(). */
    void setPosition(const qglviewer::Vec& pos) { frame.setPosition(pos); };

    void setOrientation(const qglviewer::Quaternion& q);
    /*! Same as setOrientation(), but with float parameters. */
    void setOrientation(float q0, float q1, float q2, float q3) { setOrientation(qglviewer::Quaternion(q0,q1,q2,q3)); };
    void setOrientation(float theta, float phi);

    void setCameraFromProjectionMatrix(const float* const m);

  public:    
    void getPosition(float& x, float& y, float& z) const;
    /*! Returns the camera position (the eye), defined in the world coordinate system.
     * Actually calls worldCoordinatesOf(). Use setPosition() to set the camera position. */
    qglviewer::Vec position() const { return frame.position(); };

    void getOrientation(float& q0, float& q1, float& q2, float& q3) const;
    /*! Returns the camera current orientation, defined in the world coordinate system.
     * Actually calls \p frame.orientation(). Use setOrientation() or lookAt() to set the camera orientation. */
    qglviewer::Quaternion orientation() const { return frame.orientation(); };
    //@}
    
    /*! @name Revolve Around Point (in REVOLVE mode)
     */
    //@{
 public slots:
    void setRevolveAroundPoint(const qglviewer::Vec& rap);
    /*! float version of setRevolveAroundPoint(). */
    void setRevolveAroundPoint(float x, float y, float z) { setRevolveAroundPoint(qglviewer::Vec(x,y,z)); };

    void setRevolveAroundPointFromPixel(int x, int y);
    /*! See setRevolveAroundPoint() and sceneCenter(). Used by the ManipulatedFrame. */
    void setRevolveAroundPointToSceneCenter() { setRevolveAroundPoint(sceneCenter()); };
  public:    
    /*! When the camera is in REVOLVE mode (see mode()), this is the point the camera
     * revolves around. Default value is the sceneCenter().
     * Camera mode can be changed with toggleMode() or with setMode().
     \attention setSceneCenter() changes this value. */
    qglviewer::Vec revolveAroundPoint() const { return frame.referenceFrame()->position(); };
    void getRevolveAroundPoint(float& x, float& y, float& z) const;
    //@}

    
    /*! @name Camera positionning tools
     */
    //@{
    /*! Returns a pointer to the Constraint that is currently attached to the camera. See setConstraint(). */
    Constraint* constraint() { return frame.constraint(); };
public slots:
    void lookAt(const qglviewer::Vec& target);
    /*! Same as lookAt(), but with float parameters. */
    void lookAt(float x, float y, float z) { lookAt(qglviewer::Vec(x,y,z)); };

    /*! Same as fitSphere(), but with float parameters. */
    void fitSphere(float x, float y, float z, float r) { fitSphere(qglviewer::Vec(x,y,z), r); };
    void fitSphere(const qglviewer::Vec& center, float radius);

    void setUpVector(const qglviewer::Vec& up, bool noMove = true);
    /*! Same as setUpVector(), but with float parameters. */
    void setUpVector(float x, float y, float z, bool noMove = true) { setUpVector(qglviewer::Vec(x,y,z), noMove); };
    
    void fitBBox(const qglviewer::Vec& m, const qglviewer::Vec& M);

    void showEntireScene();
    void centerScene();
    
    void setConstraint(Constraint* c);

    void zoomOnPixel(int x, int y);
    void zoomOutFromPixel(int x, int y);
    
  public:
    qglviewer::Vec upVector() const;
    void getUpVector(float& x, float& y, float& z) const;

    /*! Returns the normalized view direction of the camera (negative Z axis), defined in the world coordinate system. */
    qglviewer::Vec viewDirection() const { return frame.inverseTransformOf(Vec(0.0, 0.0, -1.0)); };
    //@}

    
    /*! @name Camera Paths

      A camera path is defined by a KeyFrameInterpolator (see setKeyFrameInterpolator()). You can
      interactively addKeyFrame() to a given path, mainly using the default QGLViewer shortcut:
      Alt-F[1-12] (see QGLViewer::setKeyFrameKey()). Use playKeyFramePath() to make the camera
      follow the path (default shortcut = F[1-12]). */
    //@{
public:
    /*! Returns the number of camera paths that can be defined. This is the maximum number of
      KeyFrameInterpolator that can be saved (index range is 0..\p nbPaths()-1), but some of them
      may not be \e defined (keyFrameInterpolator() will return \c NULL). */
    int nbPaths() const { return nbPaths_; };

    qglviewer::KeyFrameInterpolator* keyFrameInterpolator(unsigned short i) const;

public slots:
    // Removed in version 1.3 
    void saveCurrentPOV(unsigned short i = 0);
    void restorePOV(unsigned short i = 0);

    void setKeyFrameInterpolator(unsigned short i, qglviewer::KeyFrameInterpolator* const kfi);

    void addKeyFrame(int i);
    void playKeyFramePath(int i);
    void deleteKeyFramePath(int i);

    void drawAllPaths();
    //@}

protected:  
    virtual void loadProjectionMatrix(bool reset=true) const;
    
    virtual void loadProjectionMatrixStereo(bool leftBuffer=true) const;
    virtual void loadModelViewMatrix(bool reset=true) const;
     
    virtual void loadModelViewMatrixStereo(bool leftBuffer=true) const;

    virtual void mouseDoubleClickEvent(QMouseEvent *);
    virtual void adaptFrameState(qglviewer::Frame* const fr, QMouseEvent* const) const;

    /*! Returns the scene radius used by the camera. Used to set the camera near and far planes (see zNear() and zFar()).
     * This value is updated by the viewer through setSceneRadius(). */
    float sceneRadius() const { return sceneRadius_; };

    void getSceneCenter(float& x, float& y, float& z) const;
    /*! Returns the position of the scene center. This value is set by the QGLViewer.
     * Default is 0,0,0 (world origin). */
    qglviewer::Vec sceneCenter() const { return sceneCenter_; };

    void computeModelViewMatrix() const;

protected slots :
    /*! Update a copy of the viewer window dimension. Called by the QGLViewer when it is resized.*/
    void setWindowWidthAndHeight(int w, int h);
   
  private:
    // Copy constructor and opertor= are declared private and undefined
    // Prevents everyone from trying to use them
    Camera(const Camera& v);
    Camera& operator=(const Camera& v);

    // C a m e r a   p a r a m e t e r s
    int screenWidth_, screenHeight_;  // size of the window, in pixels
    float fieldOfView_;
    float sceneRadius_;
    float zNearCoef_;
    qglviewer::Vec sceneCenter_;
    Type type_; // PERSPECTIVE or ORTHO
    mutable GLdouble modelViewMatrix_[16]; // Buffered model view matrix.

    // S t e r e o   p a r a m e t e r s 
    float IODist_;            // inter-ocular distance
    float distanceToScreen_;  // in meters
    float distanceToZeroParallaxPlane_;  // in scene units
    float physicalScreenWidth_;       // in meters

    // P o i n t s   o f   V i e w s   a n d   K e y F r a m e s 
    enum { nbPaths_ = 12 };
    qglviewer::KeyFrameInterpolator* kfi_[nbPaths_];
    unsigned short lastActivatedKFNumber_; // To prevent double path interpolation
    qglviewer::Frame* dummyFrame;  // Remove in QGLViewer version 1.3
 private slots:
    void updateFromKFI();  // Remove in QGLViewer version 1.3
  };


////////////////////// Deprecated functions /////////////////
// #define __WHERE__ "In file "<<__FILE__<<", line n"<<__LINE__<< ", use "
// #define orientationAxisAngle(x,y,z,a) 	{ std::cout << __WHERE__ << "getOrientationAxisAngle()." << std::endl; exit(0); }

} // namespace qglviewer

#endif // QGLVIEWER_CAMERA_H
