// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// QGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// QGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with QGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



// Ce programme est libre, vous pouvez le redistribuer et/ou le modifier
// selon les termes de la Licence Publique Gnrale GNU publie par la Free
// Software Foundation (version 2 ou bien toute autre version ultrieure
// choisie par vous).

// Ce programme est distribu car potentiellement utile, mais SANS AUCUNE
// GARANTIE, ni explicite ni implicite, y compris les garanties de
// commercialisation ou d'adaptation dans un but spcifique. Reportez-vous
//  la Licence Publique Gnrale GNU pour plus de dtails.

// Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en
// mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis. 

#ifndef QGLVIEWER_CONSTRAINT_H
#define QGLVIEWER_CONSTRAINT_H

#include "vec.h"
#include "quaternion.h"

namespace qglviewer {
  class Frame;
  class Camera;

  /*!
   * \brief A base class for Frame constraints.
   *
   * The different constraints that can be applied to a Frame to limit its motion must all derive
   * from this class. A frame can have its constrain defined with Frame::setConstraint().
   *
   * constrainTranslation() and constrainRotation() can be overloaded to specify the constraint
   * behavior. They take the motion and the applied frame as parameters. The translation (resp.
   * rotation) should be modified according to the constraint.
   *
   * A frame can be moved with constraints using Frame::translate() and Frame::rotate(). It
   * will result in the following pseudo-code for an imposed translation \p T and rotation \p Q :
   \code
   constraint()->constrainTranslation( T , this );
   t += T;
   constraint()->constrainRotation   ( Q , this );
   q *= Q;
   \endcode
   *
   * Default behavior (Constraint base class) implements no constraint.
   * \attention Frame::setTranslation(), Frame::setRotation() and similar functions will actually \e
   * set the frame position and orientation, and the constraint will not be used. */
  class Constraint
  {
  public:
    /*! Virtual empty destructor. */
    virtual ~Constraint() {};

    /*! Default un-constrained translation function. Desired translation is expressed in \e local frame
      coordinate system. Try the \p constrainedFrame and <a href="../examples/constrainedCamera.html">constrainedCamera</a> examples for an
      illustration.

      Overload this function in your own Constraint class to define a new translation constraint.
      The frame which is applied a constrained translation is given as a parameter. It is not
      defined \c const, but you should refrain from directly changing its value in the
      constraint. */
    virtual void constrainTranslation(qglviewer::Vec&, Frame* const) {};

    /*! Default un-constrained rotation function. Desired rotation is expressed in \e local frame
      coordinate system. Try the \p constrainedFrame and <a href="../examples/constrainedCamera.html">constrainedCamera</a> examples for an
      illustration.

      Overload this function in your own Constraint class to define a new rotationnal constraint.
      The frame which is applied a constrained rotation is given as a parameter. It is not defined
      \c const, but you should refrain from directly changing its value in the constraint. */
    virtual void constrainRotation(qglviewer::Quaternion& , Frame* const) {};
  };

  /*!
   * \brief An abstract class for Frame constraints that use an axis or a plane.
   *
   * This class defines and implements a translation and a rotation constraint direction and type.
   * The type is defined by the Type() enum which could be extended in derived classes to define new
   * constraint types.
   *
   * Different implementations of this class are illustrated in the
   <a href="../examples/constrainedCamera.html">contrainedCamera</a> and
   <a href="../examples/constrainedFrame.html">constrainedFrame</a> examples.

   \attention The Axis rotationnal constraint is implemented by projecting the rotation axis
   drection of the constrained axis. The constrain is hence satisfied, but unpleasant behavior may
   appear when a mouse ManipulatedFrame is used to move the frame (try the above examples to see
   what I mean). A solution would be to directly measure the rotation angle in screen coordinates,
   but that would imply to know the Camera associated to the viewer, so that we can compute the
   projected coordinates of the rotation center. This is the way the Camera screen Z axis rotation
   is implemented (see the <a href="../mouse.html">mouse page</a> for details). However, we chose \e
   not to fix this problem, as it would imply an extra Camera pointer in all the AxisPlaneConstraint
   derived classes (which the user should have to update depending on the used viewer in a
   multi-viewer application).
  */
  class AxisPlaneConstraint : public Constraint
  {
  public:
    AxisPlaneConstraint();
    /*! Virtual empty destructor. */
    virtual ~AxisPlaneConstraint() {};

    /*! The different types of translation and rotation constraints that are available. New derived
     * classes can introduce their own enum for specific constraints.
     *
     * Use setTranslationConstraint() and setRotationConstraint() to set constraints,
     * translationConstraintType() and rotationConstraintType() to get the current constraint type.
     
     \note See the warning section of the Detailed Description section for limitations on the AXIS
     rotationnal constraint.
    */
    enum Type { FREE, PLANE, AXIS, FORBIDDEN };

    /*! @name Translation constraints */
    //@{   
    /*! Default un-constrained translation function. Desired translation is expressed in local frame
      coordinate system. Try the <a href="../examples/constrainedFrame.html">constrainedFrame</a>
      and <a href="../examples/constrainedCamera.html">constrainedCamera</a> examples for an
      illustration. */
    virtual void constrainTranslation(qglviewer::Vec&, Frame* const)=0;
    
    void setTranslationConstraint (const Type c, const qglviewer::Vec& dir);
    void setTranslationConstraint (const Type c, const float x, const float y, const float z);
    /*! Set the Type() of the translation constraint. See setTranslationConstraint(). */
    void setTranslationConstraintType (const Type c) { translationConstraint_ = c; };
    void setTranslationConstraintDir(const qglviewer::Vec& dir);
  
    /*! Returns the current translation constraint. Result is of type Type(). The
     * translation constraint is set with setTranslationConstraint(). */
    Type translationConstraintType() const { return translationConstraint_; };
 
    /*! Same as getTransConstraintDir() but result is returned as floats. */
    void getTranslationConstraintDir(float& x, float& y, float& z) const;
    /*! Returns the current direction of the translation constraint. If translationConstraintType()
     * is FREE or FORBIDDEN, this value is undefined. \p dir is the plane normal or the axis
     * direction, depending on the constraint type. It is expressed in different coordinate system,
     * depending on the actual constraint class (camera for CameraConstraint, local for
     * LocalConstraint, world for WorldConstraint, etc.) This value can be changed with
     * setTranslationConstraintDir(). */
    qglviewer::Vec translationConstraintDir() const { return translationConstraintDir_; };
    //@}
    
    /*! @name Rotation constraints */
    //@{   
    /*! Default un-constrained rotation function. Desired rotation is expressed in local frame
      coordinate system. Try the <a href="../examples/constrainedFrame.html">constrainedFrame</a>
      and <a href="../examples/constrainedCamera.html">constrainedCamera</a> examples for an
      illustration. */
    virtual void constrainRotation(qglviewer::Quaternion& , Frame* const)=0;
      
    void setRotationConstraint (const Type c, const qglviewer::Vec& dir);
    void setRotationConstraint (const Type c, const float x, const float y, const float z);
    /*! Set the Type() of the rotation constraint. See setRotationConstraint(). */
    void setRotationConstraintType (const Type c) { rotationConstraint_ = c; };
    void setRotationConstraintDir(const qglviewer::Vec& dir);

    /*! Returns the current rotation constraint. Result is of type Type(). The rotation
     * constraint is set with setRotationConstraint(). */
    Type rotationConstraintType() const { return rotationConstraint_; };

    /*! Same as getRotConstraintDir() but result is returned as floats. */
    void getRotationConstraintDir(float& x, float& y, float& z) const;
    /*! Returns the current direction of the rotation constraint. If rotationConstraintType() is
     * FREE or FORBIDDEN, this value is undefined. \p dir is the plane normal or the axis direction,
     * depending on the constraint. It is expressed in a different coordinate system depending on
     * the derived class (WorldConstraint, CameraConstraint()...). This value can be changed with
     * setRotationConstraint(). */
    qglviewer::Vec rotationConstraintDir() const { return rotationConstraintDir_; };
    //@}
  
  private:
    /*! A constraint Type(), used for translation constraint specification. */
    Type translationConstraint_;
    /*! A constraint Type(), used for rotation constraint specification. */
    Type rotationConstraint_;

    /*! A vector, used for translation constraints (axis direction or plane normal). */
    qglviewer::Vec translationConstraintDir_;
    /*! A vector, used for rotation constraints (axis direction or plane normal). */
    qglviewer::Vec rotationConstraintDir_;
  };

  
  /*!
   * \brief A translation/rotation Constraint defined with respect to the local frame coordinate system.
   */
  class LocalConstraint : public AxisPlaneConstraint
  {
  public:
    /*! Virtual empty destructor. */
    virtual ~LocalConstraint() {};

    virtual void constrainTranslation(qglviewer::Vec&        , Frame* const);
    virtual void constrainRotation   (qglviewer::Quaternion& , Frame* const);
  };


  
  /*!
   * \brief A translation/rotation Constraint defined with respect to the world coordinate system.
   */
  class WorldConstraint : public AxisPlaneConstraint
  {
  public:
    /*! Virtual empty destructor. */
    virtual ~WorldConstraint() {};

    virtual void constrainTranslation(qglviewer::Vec&        , Frame* const);
    virtual void constrainRotation   (qglviewer::Quaternion& , Frame* const);
  };

  
  /*!
   * \brief A translation/rotation Constraint defined with respect to the camera coordinate system.
   */
  class CameraConstraint : public AxisPlaneConstraint
  {
  public:
    explicit CameraConstraint(const qglviewer::Camera& cam);
    /*! Virtual empty destructor. */
    virtual ~CameraConstraint() {};

    virtual void constrainTranslation(qglviewer::Vec&        , Frame* const);
    virtual void constrainRotation   (qglviewer::Quaternion& , Frame* const);

    /*! The camera which frame is used to define the constraints direction, set with the constructor. */
    const Camera& camera;
  };
 
} // namespace qglviewer

#endif // QGLVIEWER_CONSTRAINT_H
