// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// QGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// QGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with QGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



// Ce programme est libre, vous pouvez le redistribuer et/ou le modifier
// selon les termes de la Licence Publique Gnrale GNU publie par la Free
// Software Foundation (version 2 ou bien toute autre version ultrieure
// choisie par vous).

// Ce programme est distribu car potentiellement utile, mais SANS AUCUNE
// GARANTIE, ni explicite ni implicite, y compris les garanties de
// commercialisation ou d'adaptation dans un but spcifique. Reportez-vous
//  la Licence Publique Gnrale GNU pour plus de dtails.

// Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en
// mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis. 

#ifndef QGLVIEWER_FRAME_H
#define QGLVIEWER_FRAME_H

#include <qobject.h>
#include <qstring.h>

#include "vec.h"
#include "quaternion.h"
#include "constraint.h"
// #include "GL/gl.h" is now included in config.h for ease of configuration


namespace qglviewer {
  /*!
   * \brief A frame in 3D space (position and orientation), with possible displacement constraints.
   *
   * The position and orientation can be set and restored. Constraints can be attached to the frame
   * in order to limit its displacements. A frame can be rotated and translated, with displacement
   * constraints satisfied. */
  class Frame : public QObject
  {
    
    Q_OBJECT

  public:
    Frame();
    Frame(const qglviewer::Vec& pos, const qglviewer::Quaternion& orien);
    
    /*! Virtual destructor. Empty. */
    virtual ~Frame() {};

    Frame& operator=(const Frame& f);

    virtual QDomElement domElement(const QString& name, QDomDocument& doc) const;
    virtual void initFromDOMElement(const QDomElement& de);

    const GLdouble* matrix() const;
    void getMatrix(GLdouble m[4][4]) const;
    
    const GLdouble* worldMatrix() const;
    void getWorldMatrix(GLdouble m[4][4]) const;

    // const float* transposedMatrix() const;
    // const float* inverseMatrix() const; todo
    void setFromMatrix(const float m[4][4]);

    signals:
    /*! This signal is emitted whenever the position or the orientation of the Frame is changed. Connect this
      signal to any object that must be notified.
      \code
      QObject::connect(myFrame, SIGNAL(modified()), myObject, SLOT(update()));
      \endcode

      This signal can also be connected to \e all the viewers using QGLViewer::connectSignalToAllViewers().

      \attention Note that connecting this signal to an updateGL() function will prevent you from
      modifying the frame inside your QGLViewer::draw() function. It will in fact create an infinite
      loop.
    */
    void modified();
    
  public:
    /*! @name Composition of frames
     * A frame is always defined with respect to a given coordinate system. Default is the
     * world coordinate system, corresponding to a \p NULL referenceFrame().
     *
     * Setting the reference frame to non \p NULL will define the frame with respect to this reference
     * frame. Hierarchical frame structures can then be represented in a tree, whose root would be
     * the world coordinate system (\p NULL). */
    //@{
    /*! Returns the current reference frame, defined with setReferenceFrame().

      The frame is "child" of the referenceFrame(), meaning that its position and orientation are
      defined with respect to the referenceFrame(). A \c NULL referenceFrame() (default value) means
      that the frame is defined is the world coordinate system. Chaining frames like this allows you
      to easily create a hierarchy of frames.

      Convenient functions allow you to convert 3D coordinates from one frame to an other :
      coordinatesOf(), localCoordinatesOf(), coordinatesOfIn() and their inverse functions.

      Vectors can also be converted from one frame to an other using transformOf(), transformOfIn,
      localTransformOf() and the inverse functions.

      A warning is printed if setting \p fr as a reference frame would create a loop in the frame
      hierarchy (see settingAsReferenceFrameWillCreateALoop()), as this hierarchy needs to be a
      tree. */
    Frame* referenceFrame() const { return referenceFrame_; };
    void setReferenceFrame(Frame* const fr);

    bool settingAsReferenceFrameWillCreateALoop(Frame* const fr);
    //@}

    /*! @name Local frame translation and rotation
     * Methods to set and get the frame translation and rotation (defined with respect to the reference frame).
     *
     * A frame is defined by its translation with respect to its reference frame, and \e then by a
     * rotation of the coordinate system around the new translated origin.
     * These values (translation and then rotation) are hence defined with respect to the local
     * reference frame. Use position() and orientation() to get the \e world coordinates of the frame. */
    //@{
    void setTranslation(const float x, const float y, const float z);
    /*! Set the current translation of the frame. position() returns the current frame position.
      Emits the modified() signal. */
    void setTranslation(const qglviewer::Vec& t) { t_ = t; emit modified(); };

    void setRotation(const float x, const float y, const float z, const float a);
    /*! Set the current rotation quaternion. See rotation() and the different Quaternion()
      constructors. Emits the modified() signal. */
    void setRotation(const qglviewer::Quaternion& q) { q_ = q; emit modified(); };
    void setRotationAxisAngle(const float x, const float y, const float z, const float a);

    void setTranslationAndRotation(const qglviewer::Vec& t, const qglviewer::Quaternion& q);

    void getTranslation(float& x, float& y, float& z) const;
    /*! Returns the current frame position. See setTranslation(). */
    qglviewer::Vec translation() const { return t_; };

    void getRotation(float& x, float& y, float& z, float& a) const;
    /*! Returns the current quaternion orientation. See setRotation(). */
    qglviewer::Quaternion rotation() const { return q_; };

    void translate(qglviewer::Vec& t);
    void translate(const qglviewer::Vec& t);
    void translate(float t[3]);
    void translate(float& x, float& y, float& z);
    void translate(const float x, const float y, const float z);
    
    void rotate(qglviewer::Quaternion& q);
    void rotate(const qglviewer::Quaternion& q);
    void rotate(float q[4]);
    void rotate(float& x, float& y, float& z, float& a);
    void rotate(const float x, const float y, const float z, const float a);
    //@}


    /*! @name World Coordinate position and orientation
     * Methods to set and restore the frame position and orientation (with respect to the \e world coordinate system).
     *
     * \p position is the absolute position of the origin of the frame, and
     * \p orientation \e then defines its orientation (as a rotation around the new translated frame origin).
     *
     * Use translation() and rotation() to get the \e local (defined with respect to the
     * referenceFrame) frame characteristics. */
    //@{
    void setPosition(const float x, const float y, const float z);
    void setPosition(const qglviewer::Vec& t);
    
    void setOrientation(const qglviewer::Quaternion& q);
    void setOrientation(const float x, const float y, const float z, const float a);
    void setOrientationAxisAngle(const float x, const float y, const float z, const float a);

    void setPositionAndOrientation(const qglviewer::Vec& t, const qglviewer::Quaternion& q);
    
    /*! Returns the current frame origin position (defined in the world coordinate system). See setPosition(). */
    qglviewer::Vec position() const { return inverseCoordinatesOf(Vec(0.0,0.0,0.0)); };
    void getPosition(float& x, float& y, float& z) const;

    qglviewer::Quaternion orientation() const;
    void getOrientation(float& x, float& y, float& z, float& a) const;
    //@}

    /*! @name Coordinate system transformation of 3D points
     *
     * A frame is as a new coordinate system, defined with respect to a reference frame
     * (the world coordinate system by default, see the "Composition of frame" section).
     *
     * The coordinatesOf() (resp. inverseCoordinatesOf()) functions transform a 3D point from (resp.
     * to) the world coordinates system. This section defines the 3D \e point transformation functions.
     * See the Coordinate system transformation of vectors below for the transformation of 3D
     * \e vectors. The difference between the two sets of functions is simple : for vectors,
     * only the rotational part of the transformations is taken into account, while translation is
     * also considered for 3D points.
     *
     * When \e local is prepended to the names of the functions, the functions
     * simply transform from (and to) the reference frame.
     *
     * When \e In (resp. From) is appended to the names, the functions transform
     * from (and to) the frame that is given as an argument. The frame does not need to be in the
     * same branch or the hierarchical tree, and can be NULL (the world coordinates system).
     *
     * Combining any of these functions with its inverse (in any order) leads to the identity. */
    //@{
    qglviewer::Vec coordinatesOf(const qglviewer::Vec& src) const;
    qglviewer::Vec inverseCoordinatesOf(const qglviewer::Vec& src) const;
    qglviewer::Vec localCoordinatesOf(const qglviewer::Vec& src) const;
    qglviewer::Vec localInverseCoordinatesOf(const qglviewer::Vec& src) const;
    qglviewer::Vec coordinatesOfIn(const qglviewer::Vec& src, const Frame* in) const;
    qglviewer::Vec coordinatesOfFrom(const qglviewer::Vec& src, const Frame* from) const;
    
    void getCoordinatesOf(const float src[3], float res[3]) const;
    void getInverseCoordinatesOf(const float src[3], float res[3]) const;
    void getLocalCoordinatesOf(const float src[3], float res[3]) const;
    void getLocalInverseCoordinatesOf(const float src[3], float res[3]) const;
    void getCoordinatesOfIn(const float src[3], float res[3], const Frame* in) const;
    void getCoordinatesOfFrom(const float src[3], float res[3], const Frame* from) const;
    //@}

    /*! @name Coordinate system transformation of vectors
     *
     * A frame is as a new coordinate system, defined with respect to a reference frame
     * (the world coordinate system by default, see the "Composition of frame" section).
     *
     * The transformOf() (resp. inverseTransformOf()) functions transform a 3D \e vector from (resp.
     * to) the world coordinates system. This section defines the 3D \e vector transformation
     * functions. See the Coordinate system transformation of 3D points above for the transformation
     * of 3D \e points. The difference between the two sets of functions is simple : for vectors,
     * only the rotational part of the transformations is taken into account, while translation is
     * also considered for 3D points.
     *
     * The length of the resulting transformed vector is identical to the one of the source vector
     * for all the described functions.
     *
     * When \e local is prepended to the names of the functions, the functions
     * simply transform from (and to) the reference frame.
     *
     * When \e In (resp. From) is appended to the names, the functions transform
     * from (and to) the frame that is given as an argument. The frame does not need to be in the
     * same branch or the hierarchical tree, and can be NULL (the world coordinates system).
     *
     * Combining any of these functions with its inverse (in any order) leads to the identity. */
    //@{
    qglviewer::Vec transformOf(const qglviewer::Vec& src) const;
    qglviewer::Vec inverseTransformOf(const qglviewer::Vec& src) const;
    qglviewer::Vec localTransformOf(const qglviewer::Vec& src) const;
    qglviewer::Vec localInverseTransformOf(const qglviewer::Vec& src) const;
    qglviewer::Vec transformOfIn(const qglviewer::Vec& src, const Frame* in) const;
    qglviewer::Vec transformOfFrom(const qglviewer::Vec& src, const Frame* from) const;
    
    void getTransformOf(const float src[3], float res[3]) const;
    void getInverseTransformOf(const float src[3], float res[3]) const;
    void getLocalTransformOf(const float src[3], float res[3]) const;
    void getLocalInverseTransformOf(const float src[3], float res[3]) const;
    void getTransformOfIn(const float src[3], float res[3], const Frame* in) const;
    void getTransformOfFrom(const float src[3], float res[3], const Frame* from) const;

    // Deprecated functions
    qglviewer::Vec orientationOf(const qglviewer::Vec& src) const;
    qglviewer::Vec inverseOrientationOf(const qglviewer::Vec& src) const;
    //@}

    /*! @name Constraints on the displacement
     *
     * A constraint can be attached to the frame to limit its displacement. The translate() and rotate() function will satisfy this constraint.
     * Default value is a non-constraint constraint. See the Constraint() and derived class documentation.
     */
    //@{
    /*! Returns the current constraint. It is of type Constraint, or one of its derived classes. Consider using dynamic_cast. */
    Constraint* constraint() const { return constraint_; }
    void setConstraint(Constraint* c);
    //@}
    
  private:
    // Copy constructor is declared private and undefined.
    // Prevents everyone from trying to use it.
    Frame(const Frame& f);

    //! P o s i t i o n   a n d   o r i e n t a t i o n
    qglviewer::Vec t_;
    qglviewer::Quaternion q_;

    // C o n s t r a i n t s
    Constraint* constraint_;
    static Constraint* const noConstraint_;
    
    // F r a m e   c o m p o s i t i o n
    Frame* referenceFrame_;    
  };
  
} // namespace qglviewer

#endif // QGLVIEWER_FRAME_H
