// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// QGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// QGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with QGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



// Ce programme est libre, vous pouvez le redistribuer et/ou le modifier
// selon les termes de la Licence Publique Gnrale GNU publie par la Free
// Software Foundation (version 2 ou bien toute autre version ultrieure
// choisie par vous).

// Ce programme est distribu car potentiellement utile, mais SANS AUCUNE
// GARANTIE, ni explicite ni implicite, y compris les garanties de
// commercialisation ou d'adaptation dans un but spcifique. Reportez-vous
//  la Licence Publique Gnrale GNU pour plus de dtails.

// Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en
// mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis. 

#ifndef QGLVIEWER_KEY_FRAME_INTERPOLATOR_H
#define QGLVIEWER_KEY_FRAME_INTERPOLATOR_H

#include <qobject.h>
#include <qtimer.h>
#include <qvaluevector.h>

#include "frame.h"

namespace qglviewer {
  /*!
   \brief A key frame interpolator.
   
   KeyFrameInterpolator implements a classical interpolation of key Frames, using a smooth
   Catmull-Rom based interpolation.

   Use addKeyFrame() to define the keyframes and then call startInterpolation(). The Frame
   associated with the interpolator (using the constructor or with setFrame()) will then be updated over
   time.

   The interpolation can loop (setLoopInterpolation()) and the keyFrame path can be closed
   (setClosedPath()). The current interpolation time (interpolationTime()) and speed
   (interpolationSpeed()) can also be tuned.

   The keyFrames are defined by a Frame and a time, expressed in seconds. The time has to be
   monotonously increasing over keyFrames.

   See the <a href="../examples/keyFrames.html">keyFrames</a> example for an illustration of the class.
   
   \todo : insertKeyFrames, deleteKeyFrame, replaceKeyFrame...
  */
  class KeyFrameInterpolator : public QObject
  {
    Q_OBJECT
    
  public:
    KeyFrameInterpolator(qglviewer::Frame* fr=NULL);
    /*! Virtual destructor. Empty. */
    virtual ~KeyFrameInterpolator() {};

    virtual QDomElement domElement(const QString& name, QDomDocument& doc) const;
    virtual void initFromDOMElement(const QDomElement& de);

    /*! @name Key Frames
     */
    //@{
    /*! Set the frame that will be associated with the KeyFrameInterpolator. Current frame is in frame(). */
    void setFrame(qglviewer::Frame* fr) { frame_ = fr; };
    /*! The frame that is associated and modified by the KeyFrameInterpolator. Set using setFrame()
      or with the KeyFrameInterpolator() constructor. */
    qglviewer::Frame* frame() const { return frame_; };

    void addKeyFrame(const qglviewer::Frame& fr);
    void addKeyFrame(const qglviewer::Frame& fr, const float time);

    void addKeyFrame(const qglviewer::Frame* fr);
    void addKeyFrame(const qglviewer::Frame* fr, const float time);

    void deletePath();

    /*! Number of keyFrames used in the interpolation. Use addKeyFrame() to add new keyFrames. */
    int numberOfKeyFrames() const { return keyFrame_.size(); };
    
    qglviewer::Vec keyFramePosition(const int i) const;
    qglviewer::Quaternion keyFrameOrientation(const int i) const;
    /*! Returns the time associated with the keyframe number \p i. See also keyFramePosition(),
      keyFrameOrientation() and addKeyFrame(). */
    float keyFrameTime(const int i) const { return keyFrame_[i].time_; };
    //@}
  
    /*! @name Interpolation parameters
     */
    //@{
    /*! Current interpolation time, updated when interpolationIsStarted() is true. Can be set
      directly with setInterpolationTime() or interpolateAtTime(). */
    float interpolationTime() const { return interpolationTime_; };
    /*! Set the current interpolation time. See also interpolationTime().

      \attention The frame() is not affected by this method. Use this function to define the
     starting time of a future interpolation (see startInterpolation()). Use interpolateAtTime() to
     actually interpolate at a given time. */
    void setInterpolationTime(const float time) { interpolationTime_ = time; };

    /*! Returns the current interpolation speed. Default value is 1.0, which means that the times
      associated with the keyFrames (see keyFrameTime() and addKeyFrame()), expressed in seconds,
      will be respected during the interpolation (see startInterpolation()), provided that your
      display is fast enough.
      
      A negative value will result in a reverse interpolation of the keyFrames. */
    float interpolationSpeed() const { return interpolationSpeed_; };
    /*! Set the interpolation speed. Negative values are possible. See interpolationSpeed(). */
    void setInterpolationSpeed(const float speed) { interpolationSpeed_ = speed; };

    /*! Returns the current interpolation period, expressed in milliseconds. This period will be
      added to the interpolationTime() at each update(), when interpolationIsStarted() is \p true. */
    int interpolationPeriod() const { return period_; };
    /*! Set the interpolationPeriod(). */
    void setInterpolationPeriod(const int msecs) { period_ = msecs; };
    
    /*! Whether or not (default) the interpolation will be played in (infinite) loop when
      startInterpolation() is called (until stopInterpolation() is called). When \p false, the
      interpolation will stop when interpolationTime() reaches the last keyFrames associated time
      (see keyFrameTime()).

      Use setClosedPath() to create a closed path with the keyFrames. */
    bool loopInterpolation() const { return loopInterpolation_; };
    /*! Set the loopInterpolation() value. */
    void setLoopInterpolation(const bool loop=true) { loopInterpolation_ = loop; };

    /*! Whether or not (default) the path defined by the keyFrames is a closed loop. When \p true,
      the last and the first KeyFrame are linked by a new spline segment. 
      \todo Not yet implemented and specified.
      
      Use setLoopInterpolation() to create a continuous animation over the entire path. */
    bool closedPath() const { return closedPath_; };
    /*! Set the closedPath() value. */
    void setClosedPath(const bool closed=true) { closedPath_ = closed; };
    //@}

    
    /*! @name Interpolation
     */
    //@{
    virtual void interpolateAtTime(const float time);
    void startInterpolation(const int msecs = -1);
    void stopInterpolation();
    void resetInterpolation();

    /*! Returns \p true when the interpolation is being processed (after a call to
      startInterpolation()). Set to \p false by stopInterpolation(). See also
      toggleInterpolation(). */
    bool interpolationIsStarted() const { return interpolationStarted_; };
    /*! Calls startInterpolation() or stopInterpolation(), depending on interpolationIsStarted(). */
    void toggleInterpolation() { if (interpolationIsStarted()) stopInterpolation(); else startInterpolation(); };

    //@}
    
    virtual void drawPath(const int mask=1, int nbFrames=6, const float scale=1.0);

  public slots:
    virtual void update();
      
  private slots:
    virtual void invalidateValues() { valuesAreValid_ = false; pathIsValid_ = false; };
    
  private:
    // Copy constructor and opertor= are declared private and undefined
    // Prevents everyone from trying to use them
    KeyFrameInterpolator(const KeyFrameInterpolator& kfi);
    KeyFrameInterpolator& operator=(const KeyFrameInterpolator& kfi);

    void updateCurrentKeyFrameForTime(const float time);
    void updateModifiedFrameValues();

    //! Internal private KeyFrame representation
    class KeyFrame
    {
    public :
      KeyFrame() {};
      KeyFrame(const qglviewer::Frame& fr, const float t);
      KeyFrame(const qglviewer::Frame* fr, const float t);
      void updateValues();
      qglviewer::Vec position() const;
      qglviewer::Quaternion orientation() const;
      float value_[9];
      float time_;
      const qglviewer::Frame* frame_;
    };

    void interpolate(const float alpha, const QValueVector<KeyFrame>::const_iterator kf[4],
		     qglviewer::Frame* fr);

    //! K e y F r a m e s
    QValueVector<KeyFrame> keyFrame_;
     // 4 succesive frames. interpolationTime_ is between index 1 and 2.
    QValueVector<KeyFrame>::const_iterator currentFrame_[4];
    QValueVector<qglviewer::Frame> path_;
    
    //! A s s o c i a t e d   f r a m e
    qglviewer::Frame* frame_;

    //! R y t h m
    QTimer timer_;
    int period_;
    float interpolationTime_;
    float interpolationSpeed_;
    bool interpolationStarted_;

    //! M i s c
    bool closedPath_;
    bool loopInterpolation_;
    bool pathIsValid_;
    bool valuesAreValid_;
    bool currentFrameValid_;
    // float sceneRadius_;
  };
  
} // namespace qglviewer

#endif // QGLVIEWER_KEY_FRAME_INTERPOLATOR_H
