// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// QGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// QGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with QGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



// Ce programme est libre, vous pouvez le redistribuer et/ou le modifier
// selon les termes de la Licence Publique Gnrale GNU publie par la Free
// Software Foundation (version 2 ou bien toute autre version ultrieure
// choisie par vous).

// Ce programme est distribu car potentiellement utile, mais SANS AUCUNE
// GARANTIE, ni explicite ni implicite, y compris les garanties de
// commercialisation ou d'adaptation dans un but spcifique. Reportez-vous
//  la Licence Publique Gnrale GNU pour plus de dtails.

// Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en
// mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis. 

#ifndef QGLVIEWER_MANIPULATED_FRAME_H
#define QGLVIEWER_MANIPULATED_FRAME_H

#include <qstring.h>
#include <qtimer.h>
#include <qdatetime.h>

#include "frame.h"

class QGLViewer;

namespace qglviewer {

  class Camera;
  /*!
   * \brief A class that encapsulates the behavior of a manipulated Frame.
   *
   * A manipulated Frame is a Frame that can be rotated and translated with a mouse. It is used to
   * set and move the position of a camera or of an object of the scene.
   *
   * Two frame manipulation modes are supported : REVOLVE and FLY (see Mode()). */
  class ManipulatedFrame : public Frame
  {
#ifndef DOXYGEN
    friend class Camera;
    friend class ::QGLViewer;
#endif

    Q_OBJECT

  public:
    ManipulatedFrame();
    /*! Virtual destructor. Empty. */
    virtual ~ManipulatedFrame() {};
    
    signals:
    /*! This signal is emitted whenever the frame is manipulated using the mouse. Connect this
      signal to any object that must be notified. Note that when a manipulatedFrame is attached to a
      QGLViewer (see QGLViewer::setManipulatedFrame()), this signal is automatically connected to \p
      all the viewers' updateGL() functions (see QGLViewer::connectSignalToAllViewers()).
     */
    void manipulated();
    
  public:
    virtual QDomElement domElement(const QString& name, QDomDocument& doc) const;
    virtual void initFromDOMElement(const QDomElement& de);

    /*! Mode() describes the different modes of the ManipulatedFrame, set by setMode().
     *
      In REVOLVE mode, the frame revolves around a fixed point of the scene. When used for a camera,
      this is the Camera::revolveAroundPoint. For a scene ManipulatedFrame, it is the frame origin.
      *
      * The FLY mode is made for a manipulated camera frame. The camera can then slowly move forward
      * or backward in the scene, as if it was flying.

      * See mousePressEvent() for mouse button binding details. Current Mode value is returned by mode().
      * Use toggleMode() and setMode() to change mode. Example:
      \code
      camera.setMode(qglviewer::ManipulatedFrame::FLY);
      \endcode
    */
    enum Mode { REVOLVE , FLY };

    /*! @name Intrinsic parameters
     */
    //@{    
  public slots:
    void setMode(const Mode mode);
    void toggleMode();

    /*! When ManipulatedFrame is in FLY mode (see Mode()), this value tunes the speed
     * of the displacement. Current value is given by flySpeed(). */
    void setFlySpeed(const float s) { flySpeed_ = s; };
    
    /*! Defines the influence of a mouse displacement on the frame rotation. Default value is 1.0. Defined with setRotationSensitivity(). */
    void setRotationSensitivity(const float s) { rotSensitivity_ = s; };
    /*! Defines the influence of a mouse displacement on the frame translation. Default value is
      1.0. Defined with setTranslationSensitivity().

      \note You should not have to modify this value as the frame translation is computed so that
      the frame intuitively exactly matches the mouse trajectory.

      \note When the manipulated frame is used to move a \e Camera, you may be worried when you
       are looking at a small region of your scene, which translate too fast on the screen when you move the camera.

      For a camera, it is the Camera::revolveAroundPoint() that exactly matches the mouse
      displacement. Instead of changing the translationSensitivity(), you may prefer to temporarilly
      set the Camera::RevolveAroundPoint() to a point of the small region of your scene (see the <a
      href="../mouse.html">mouse page</a> for the mouse binding). */
    void setTranslationSensitivity(const float s) { transSensitivity_ = s; };
    
  public:
    /*! Returns the current ManipulatedFrame mode. Mode is set by setMode().
     * Returned values are defined by the Mode() enum. */
    Mode mode() const { return mode_; };

    /*! Returns the current fly speed (used when mode() is FLY). Set with setFlySpeed().
     * Set to 0.5% of the sceneRadius in Camera::setSceneRadius(). */
    float flySpeed() const { return flySpeed_; };

    /*! Returns the current rotationSensitivity() as defined by setRotationSensitivity(). Default value is 1.0 */
    float rotationSensitivity() const { return rotSensitivity_; };
    /*! Returns the current translationSensitivity() as defined by setTranslationSensitivity(). Default value is 1.0 */
    float translationSensitivity() const { return transSensitivity_; };
    //@}

    
    /*! @name Current state
     */
    //@{
    /*! \p True when the ManipulatedFrame is being manipulated with the mouse. Can be used to
      manually simulate the QGLViewer::fastDraw() function. Used by QGLViewer. */
    bool isManipulated() const { return action_ != NO_ACTION; };

    /*! \p True when the ManipulatedFrame is spinning (fast rotation when mouse button is released) */
    bool isSpinning() const { return spinning_; };

    /*! Return the up vector used in FLY mode. The vector is orthogonal to the horizontal plane.
      Horizontal displacement of the mouse rotates the frame around this vector. */
    qglviewer::Vec flyUpVector() const { return flyUpVector_; };
    //@}
    
  protected:
    /*! @name Internal methods for spinning - Should be moved in a spinningFrame class.
     */
    //@{
    /*! Starts the spinning of the frame, defined by spinningQuaternion(). Use stopSpinning() to stop.
      \p delay msec will separate each update of the spinning. */
    virtual void startSpinning(const unsigned int delay=0) { timer_.start(delay); spinning_ = true;  };
    /*! Stops the spinning which was initiated with startSpinning(). Spinning is initiated when the mouse is released
      with a sufficient speed in REVOLVE::Rotation mode. Clicking the mouse stops the spinning. */
    virtual void stopSpinning()  { timer_.stop();  spinning_ = false; };
    //@}

    /*! @name Mouse event handlers
     */
    //@{
    virtual void mousePressEvent(QMouseEvent *, const Camera&);
    // virtual void mouseDoubleClickEvent(QMouseEvent *) {};
    virtual void mouseReleaseEvent(QMouseEvent *, Camera&);
    virtual void mouseMoveCameraEvent(const int x, const int y, const Camera&);
    virtual void mouseMoveFrameEvent (const int x, const int y, const Camera&);
    virtual void wheelEvent(QWheelEvent *, const Camera& camera);
    //@}

  public:
    void alignWithAxis();
    
  protected slots:

    void updateFrame();
    void updateFrame_Camera();
    
  protected:
    void updateFlyUpVector();

  private:
    // Copy constructor and opertor= are declared private and undefined
    // Prevents everyone from trying to use them
    ManipulatedFrame(const ManipulatedFrame& mf);
    ManipulatedFrame& operator=(const ManipulatedFrame& mf);

    enum Action { ROTATION,ZOOM,TRANSLATION, SCREEN_ROT,SCREEN_TRANS, MOVE_FORWARD,LOOKAT,MOVE_BACK, NO_ACTION };

    // Internal atomic ManipulatedFrame methods
    void startAction(const Action m, const int x, const int y);
    void lastQuatFromDeformedBall(const int x, const int y, const float cx, const float cy, const Camera& camera);
    void lastQuatFromThetaPhi(const int x, const int y, const Camera& camera);

    // Internal state values
    Action action_;
    Mode mode_;
    float flySpeed_;                     // Speed in FLY mode, set by the camera
    qglviewer::Vec flyUpVector_;
    
    // Sensitivity
    float rotSensitivity_;
    float transSensitivity_;
    
    // Previous mouse positions.
    int prev_x,  prev_y;
    int pressX_, pressY_;

    // threshold to switch SCREEN_TRANS direction
    unsigned short threshold_;
  
    // Mouse double simple click.
    bool mouseDidNotMove_;
    bool doubleSimpleClick_[3]; // index correspond to left, middle and right button
  
    // Spinning related values
    QTime last_move_time;
    float mouseSpeed_; 
    int delay_; 
    bool spinning_;
    //bool displayManipulatedFrame; // Not used. Should display only after an immobile wait.
    qglviewer::Quaternion last_quat;
    QTimer timer_;
  };

} // namespace qglviewer

#endif // QGLVIEWER_MANIPULATEDFRAME_H
