// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// QGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// QGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with QGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



// Ce programme est libre, vous pouvez le redistribuer et/ou le modifier
// selon les termes de la Licence Publique Gnrale GNU publie par la Free
// Software Foundation (version 2 ou bien toute autre version ultrieure
// choisie par vous).

// Ce programme est distribu car potentiellement utile, mais SANS AUCUNE
// GARANTIE, ni explicite ni implicite, y compris les garanties de
// commercialisation ou d'adaptation dans un but spcifique. Reportez-vous
//  la Licence Publique Gnrale GNU pour plus de dtails.

// Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en
// mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis. 

#ifndef QGLVIEWER_QGLVIEWER_H
#define QGLVIEWER_QGLVIEWER_H

#include <qgl.h>
#include <qptrlist.h>
#include "camera.h"

/*!
 * \brief An OpenGL viewer based on QGLWidget. Provides a moving camera and snapshot facilities.
 * 
 *
 * To use it, you can derive a class from it and overload the draw() virtual method.
 * See the <a href="../examples/simpleViewer.html">simpleViewer</a> example for an introduction.
 \code
  // The Viewer class inherits from QGLViewer and overloads the draw function.
  class Viewer : public QGLViewer
  {
    Q_OBJECT // Include this if you use Qt signals/slots
  public:
    Viewer()
      : QGLViewer()
      {}
      
  protected:
    virtual void draw();
    {
    // Your OpenGL drawing code goes here...
    }
  };
 \endcode 
 *
 * An other method is to connect your drawing methods to the signals emited by the QGLViewer
 * (clean callback mechanism). See the <a href="../examples/callback.html">callback</a> example for a complete example.
 \code
 // A Scene class holds the scene data and provides a drawScene() function.
 class Scene : public QObject
 {
   Q_OBJECT // Needed in order to use the Qt signals/slots mechanism
  
   public :
     Scene(const QGLViewer* const v);
  
   public slots:
     void drawScene();
 };

 Scene::Scene(const QGLViewer* const viewer)
 : QObject()
 {
   // Connect the viewer signal to our draw function slot
   connect(viewer, SIGNAL(drawNeeded()), this, SLOT(drawScene()));
 }
 \endcode
 * The mouse and keyboard behaviors are defined with default values.
 * Redefine the shortcuts or overload the methods to define your own behavior.
 *
 * Many functionalities are implemented : snapshots savings, camera displacement with keyFramed
 * paths, manipulated frames with hierarchies and constraints. Try the numerous simple <a
 * href="../examples/index.html">examples</a> to discover the possibilities and understand how it
 * works. */
class QGLViewer : public QGLWidget
{
  Q_OBJECT        // must include this if you use Qt signals/slots
  
  // doxygen 1.2.9 seems to be messed by the Q_PROPERTY below.
#ifndef DOXYGEN
  Q_PROPERTY(QString snapshotFilename READ snapshotFilename WRITE setSnapshotFilename)
#endif
    
public:
  QGLViewer(QWidget* parent=NULL, const char* name=0,
	    const QGLWidget* shareWidget=0, WFlags f=0);
  
  QGLViewer(const QGLFormat& format, QWidget* parent=0,const char* name=0,
	    const QGLWidget* shareWidget=0, WFlags f=0);

  /*! Virtual destructor. Remove object from QGLViewer pool (see updateGLForAllViewers()) and calls saveToFile(). */
  virtual ~QGLViewer() { saveToFile(); QGLViewer::QGLViewerPool_.removeRef(this); };

  //! Temp ! Change ! Kept for backward compatibility - Removed in version 1.3
  virtual void select(QMouseEvent*) {};
  //! Temp ! Change ! select warning message - Removed in version 1.3
  void selectWarning();

  /*! @name Display of visual hints */
  //@{
public :
  /*! Returns \c true if the world axis is drawn by the viewer. Set by setDrawAxis() or toggleDrawAxis(). */
  bool axisIsDrawn() const { return drawAxis_; };
  /*! Returns \c true if a grid is drawn by the viewer. Set by setDrawGrid() or toggleDrawGrid(). */
  bool gridIsDrawn() const { return drawGrid_; };
  /*! Returns \c true if a displayFPS() is called by the viewer. Set by setDisplayFPS() or toggleDisplayFPS(). See currentFPS().*/
  bool fpsIsDisplayed() const { return displayFPS_; };
  /*! Returns \c true if drawText() and draw3DText() actually display text. Set by setEnableText() or
    toggleEnableText(). Convenient to remove all the possibly displayed text with a single key. */
  bool textIsEnabled() const { return enableText_; };
  /*! Returns \c true if the camera paths are being edited in the viewer. Set by editCameraPath() or toggleCameraPathEditor() (\c C key shortcut). */
  bool cameraPathIsEdited() const { return cameraPathIsEdited_; };
  /*! Returns \c true if the viewer is in fullScreen mode. Set by setFullScreen() or toggleFullScreen(). */
  bool isFullScreen() const { return fullScreen_; };
  /*! Returns \c true if the viewer displays in stereo. Set by setStereoDisplay() or toggleStereoDisplay(). Default value is \c false. */ 
  bool displaysInStereo() const { return stereo_; }
  

public slots:
  /*! Toggle the state of axisIsDrawn(). A wordl axis is drawn or not (\c A shortcut key). See also setDrawAxis(). */
  void toggleDrawAxis()      		{ setDrawAxis(!drawAxis_); };
  /*! Toggle the state of gridIsDrawn() flag. A XY grid is drawn or not (\c G shortcut key). See also setDrawGrid(). */
  void toggleDrawGrid()      		{ setDrawGrid(!drawGrid_); };
  /*! Toggle the state of fpsIsDisplayed(). displayFPS() is called or not (\c F shortcut key). See also setDisplayFPS() and currentFPS(). */
  void toggleDisplayFPS()    		{ setDisplayFPS(!displayFPS_); };
  /*! Toggle the state of textIsEnabled() : drawText() and draw3DText() actually draw text or not (default is yes) (\c ? shortcut key). See also setEnableText(). */
  void toggleEnableText()   		{ setEnableText(!enableText_); };
  /*! Start/stop the edition of the camera paths (\c C shortcut key). See cameraPathIsEdited(). See also editCameraPath(). */
  void toggleCameraPathEditor()		{ editCameraPath(!cameraPathIsEdited_); };
  void toggleFullScreen();
  /*! Toggle the state of displaysInStereo() (\c S shortcut key). See also setStereoDisplay(). */
  void toggleStereoDisplay()		{ setStereoDisplay(!stereo_); };

  /*! Set the state of the axisIsDrawn(). World axis is drawn or not. See also toggleDrawAxis(). */
  void setDrawAxis(bool f=true)   { drawAxis_ = f; emit axisDrawn(f); if (updateGLEnabled_) updateGL(); };
  /*! Set the state of the gridIsDrawn(). A XY grid is drawn or not. See also toggleDrawGrid(). */
  void setDrawGrid(bool f=true)	{ drawGrid_ = f; emit gridDrawn(f); if (updateGLEnabled_) updateGL(); };
  /*! Set the state of fpsIsDisplayed(). displayFPS() is called or not. See also toggleDisplayFPS() and currentFPS(). */
  void setDisplayFPS(bool f=true)	{ displayFPS_ = f; emit fpsDisplayed(f); if (updateGLEnabled_) updateGL(); };
  /*! Set the state of textIsEnabled() : drawText() and draw3DText() actually draw text or not (default is yes). See also toggleEnableText(). */
  void setEnableText(bool f=true)	{ enableText_ = f; emit textEnabled(f); if (updateGLEnabled_) updateGL(); };
  /*! Start (default, \p edit=\p true) or stop (\p edit=\p false) the edition of the camera paths. Default viewer value is \p false. See also toggleCameraPathEditor(). */
  void editCameraPath(bool edit=true)	{ cameraPathIsEdited_ = edit; emit cameraPathEdited(edit); if (updateGLEnabled_) updateGL(); };
  void setFullScreen(bool f=true);
  void setStereoDisplay(bool f=true);
  //@}


  /*! @name Display functions */
  //@{
public :
  void drawAxis(float scale=1.0);
  /*! Displays a XY grid. The calling function should set the \p modelView matrix and
    scale factor before calling this function. */
  void drawGrid() { glCallList(GridDL_); };

  virtual void startScreenCoordinatesSystem() const;
  virtual void stopScreenCoordinatesSystem() const;

protected :
  virtual void drawLight(GLenum light);
  /*! Display the frame rate in the lower left corner. The frame rate is computed as an average and
  updateGL() must be called in a loop in order to have a meaningfull value (this is the case when
  animationIsStarted()). The FPS display can be toggled with toggleDisplayFPS() or with
  setDisplayFPS(). See also currentFPS(). */ 
  virtual void displayFPS() const { drawText(10.0, 10.0, fpsString_); };

  void drawText(GLfloat x, GLfloat y, const char *text, GLfloat height=15.0) const;
  void draw3DText(const qglviewer::Vec& pos, const qglviewer::Vec& normal, const QString& string, GLfloat height=0.1) const;

private:
  /*! Eps renderer callback function. */
  virtual void drawVectorial() { makeCurrent(); paintGL(); };
#ifndef DOXYGEN
  friend void drawVectorial(void* param);
#endif
  //@}

  /*! @name Scene related parameters and methods */
  //@{
public :
  /*! The camera that is used by the viewer. See the Camera() documentation for
   * a complete description of the associated functions. */
  qglviewer::Camera camera;

  /*! Returns the aspect ratio of the GL widget.  That is the ratio between its
   * width and height (see the QGLWidget documentation). */
  float aspectRatio() const { return static_cast<float>(width())/static_cast<float>(height()); };
  
  /*! Returns the scene radius (used by the camera to set Camera::zNear() and Camera::zFar()).
   * The entire scene must be included in a sphere of radius sceneRadius(), centered on sceneCenter(). */
  float sceneRadius() const { return camera.sceneRadius(); };

  void getSceneCenter(float& x, float& y, float& z) const;
  /*! Scene should be included in a sphere centered on this point.
   * See the Camera::sceneCenter() documentation. */
  qglviewer::Vec sceneCenter() const { return camera.sceneCenter(); };

  /*! float version of revolveAroundPoint(). */
  void getRevolveAroundPoint(float& x, float& y, float& z) const { camera.getRevolveAroundPoint(x,y,z); };
  /*! When the camera is in REVOLVE mode (see mode()), this is the point the
   * camera revolves around. Default value is the sceneCenter().
   *
   * This method simply calls Camera::revolveAroundPoint(). */
  qglviewer::Vec revolveAroundPoint() const { return camera.revolveAroundPoint(); };

  /*! Returns the current manipulated frame, which will be applied the mouse motion when the
    shortcutKey() of ShortcutAction::Frame is pressed (default is Alt). Value is undefined when no
    frame has been associated with the viewer (by using setManipulatedFrame()). */
  qglviewer::ManipulatedFrame* manipulatedFrame() const { return manipulatedFrame_; };
  
  /*! Returns the current viewer frame rate, as displayed by displayFPS() (see setDisplayFPS() and
    toggleDisplayFPS()).

  This function is useful for true real-time applications that may adapt the computationnal load
  according to frame rate evolution in order to maintain a given frequency.
  
  Note that for stability reasons this value is averaged over 25 frames, and hence will only change
  every 25 frames. As Qt redraws the window only when needed, this value is meaningfull only when
  startAnimation() is called, when you use Camera::playKeyFramePath() or when the camera is
  manipulated with the mouse, so that draw() is called in a loop.
  */
  float currentFPS() { return f_p_s_; };

public slots:
  /*! Scene radius is used by the camera to scale axis and grid display, and to optimally set
   * Camera::zNear() and Camera::zFar(). The entire scene must be included in a sphere centered on
   * sceneCenter(), with radius sceneRadius(). Current value can be retrieved from sceneRadius().
   *
   * You will typically use this function in QGLViewer::init() :
   \code
   setSceneRadius(3.0);
   showEntireScene();
   \endcode
   *
   * By default the sceneRadius() is 1.0 is called in the QGLViewer constructor. */
   void setSceneRadius(float r) { camera.setSceneRadius(r); };
  
  /*! Set the center of the scene, defined in world coordinates. The entire scene should be
   * include in a sphere of radius sceneRadius(), centered on sceneCenter().
   * Default value is (0,0,0), the world coordinate system origin.
   *
   * This method should typically be called in the init() function, when the drawn
   * scene is not centered on the world origin (so that sceneRadius(), Camera::zNear() and Camera::zFar() can be
   * tightly fitted to the scene). It should normally not be modified after.
   \code
   Viewer::init()
   {
     setSceneCenter(10,20,0);
     setSceneRadius(3.0);
     showEntireScene();
   }
   \endcode
   *
   * revolveAroundPoint() has a different meaning, and is probably the value you want to change
   * while the program is running (see Camera::setRevolveAroundPoint()).
   \attention The revolveAroundPoint() is set to the sceneCenter() value by this method. */
  void setSceneCenter(const qglviewer::Vec& sc) { camera.setSceneCenter(sc); };
  /*! float version of setSceneCenter(). */
  void setSceneCenter(float x, float y, float z) { setSceneCenter(qglviewer::Vec(x,y,z)); };

  /*! Set camera so that the entire scene is visible. This method should be called in init(), once
   * the sceneRadius() and sceneCenter() have been properly set (use setSceneCenter() and
   * setSceneRadius()). Simply calls Camera::showEntireScene(). */
  void showEntireScene() { camera.showEntireScene(); if (updateGLEnabled_) updateGL(); };

  void setManipulatedFrame(qglviewer::ManipulatedFrame* fr);
  //@}

  /*! @name Snapshots */
  //@{
public:  
  /*! Returns the current snapshot filename. Set by setSnapshotFilename(). Used in saveSnapshot(). */
  const QString& snapshotFilename() const { return snapshotFilename_; };
  /*! Returns the current snapshot file format. Defined from the File Dialog window or with
    setSnapshotFormat(). Default value is "JPEG". */
  const QString& snapshotFormat() const { return snapshotFormat_; };
  
public slots:
  void saveSnapshot(bool automatic = false);
  void setSnapshotFilename(const QString& n);
  /*! Defines the snapshot file format. This format will be used by saveSnapshot() in automatic
    mode. In non-automatic mode, this will be the default format of the saveAs dialog box.

    The available formats are those handled by Qt. Classical values are "JPEG" (default
    snapshotFormat()), "PNG", "PPM, "BMP". Use the following code to get the actual list:
    \code
     QStringList formatList = QImage::outputFormatList();
     for (QStringList::Iterator it = formatList.begin(); it != formatList.end(); ++it)
       cout << (*it) << endl;
    \endcode
    
    If the library was compiled with the EPSRender option, three additional vectorial formats are
    available: "EPS", "PS" and "XFIG".

    \attention No verification is performed on the provided format validity. The next call to
    saveSnapshot() may fail if the format string is not supported by the QT QImage class.
  */
  void setSnapshotFormat(const QString& f) { snapshotFormat_ = f; };
  //@}

  /*! @name Animation */
  //@{
public:
  /*! Return true when the animation loop is started, i.e. viewer calls animate() and draw(). See
    startAnimation() and animate() */
  bool animationIsStarted() const { return animationStarted_; };
  /*! The animation loop period, in milliseconds. See startAnimation() and animate(). */
  int animationPeriod() const { return animationPeriod_; };

public slots:
  /*! Set the period (in milliseconds) between the calls to animate(). Default is 40 msecs (25
    Hz). The frame rate (keybinding 'F') will be set accordingly, provided that your animate()
    function is fast enough. */
  void setAnimationPeriod(int msecs) { animationPeriod_ = msecs; };
  virtual void startAnimation(int msecs = -1);
  virtual void stopAnimation();
  /*! Calls startAnimation() or stopAnimation(), depending on animationIsStarted(). */
  void toggleAnimationMode() { if (animationIsStarted()) stopAnimation(); else startAnimation(); };
  //@}

  
public:
  void convertClickToLine(const QMouseEvent *, float orig[3], float dir[3]) const;
  void convertClickToLine(const QMouseEvent *, qglviewer::Vec& orig, qglviewer::Vec& dir) const;


signals:
  /*! Standard message transmission. Used for progress bar for instance */
  void message(const QString&);

  /*! Signal emitted by the default QGLViewer::init() function. Connect this signal to the functions
   * that need to be called before any drawing (if any), to initialize your viewer.
   *
   * Use this signal-slot mechanism ONLY IF you don't inherit your viewer from QGLViewer
   * (in that case, simply overload the init() function). */
  void viewerInitialized();

  /*! Signal emitted by the default QGLViewer::draw() function. Connect this signal to your main
   * drawing function. preDraw() is called \e before this signal is emited and will call
   * loadProjectionMatrix() and loadModelViewMatrix() to set the viewing matrices. postDraw() will
   * be called \e after this signal is answered. pre (resp. post) draw can be tuned by adding code
   * at the beginning (resp. end) of your draw slot.
   *
   * You can check whether the camera is manipulated to call a fast draw version :
   \code
   if (viewer.camera->isManipulated())
   fastDrawMethod();
   else
   normalDrawMethod();
   \endcode
   *
   * Use this signal-slot mechanism ONLY IF you don't inherit your viewer from QGLViewer (in that
   * case, simply overload the draw() and fastDraw() function). */
  void drawNeeded();

  /*! Signal emitted by the default QGLViewer::animate() function. Connect this signal to your scene
   *  animation function.
   *
   * Use this signal-slot mechanism ONLY IF you don't inherit your viewer from QGLViewer
   * (in that case, simply overload the animate() function). */
  void animateNeeded();

  /*! Signal emitted by the default QGLViewer::help() function. Connect this signal to your own help
   *  function.
   *
   * Use this signal-slot mechanism ONLY IF you don't inherit your viewer from QGLViewer
   * (in that case, simply overload the help() function). */
  void helpRequired();

  /*! This signal is emitted whenever axisIsDrawn() changes value. This can be done with setDrawAxis() or toggleDrawAxis(). */
  void axisDrawn(bool on);
  /*! This signal is emitted whenever gridIsDrawn() changes value. This can be done with setDrawGrid() or toggleDrawGrid(). */
  void gridDrawn(bool on);
  /*! This signal is emitted whenever fpsIsDisplayed() changes value. This can be done with setDisplayFPS() or toggleDisplayFPS(). */
  void fpsDisplayed(bool on);
  /*! This signal is emitted whenever textIsEnabled() changes value. This can be done with setEnableText() or toggleEnableText(). */
  void textEnabled(bool on);
  /*! This signal is emitted whenever cameraPathIsEdited() changes value. This can be done with editCameraPath() or toggleCameraPathEditor(). */
  void cameraPathEdited(bool on);
  /*! This signal is emitted whenever isFullScreen() changes value. This can be done with setFullScreen() or toggleFullScreen(). */
  void fullScreenChanged(bool on);
  /*! This signal is emitted whenever displaysInStereo() changes value. This can be done with setStereoDisplay() or toggleStereoDisplay(). */
  void stereoChanged(bool on);

protected:
  virtual void resizeGL( int w, int h );

  virtual void initializeGL();

  /*! Initialisation of the viewer. Called in initializeGL(), before the first drawing. Can be used
   to set the camera (Camera::setPosition(), Camera::setFieldOfView(), ...) or to make the camera
   fit the scene (see Camera::showEntireScene()). OpenGL flags can also be (re)defined here.
   Default flags are set in initializeGL() but only the \c glEnable(GL_COLOR_MATERIAL) is not set
   to the OpenGL default value.

   \attention You should not call updateGL() (or any function that calls it) in this function, as
   it will create an infinite loop (this comes from the QT initializeGL() function). The different
   QGLViewer set function (setDrawGrid(), setDrawAxis()...) are protected against this problem and can
   safely be called.

   \attention This function should only be used for viewer dependent initializations (camera, openGL
   state). Other data initialisations (program parameters parsing, scene dependent
   initializations...) should be made in a separate function, called by your \p main or any other
   initialization function. On the other hand, all the openGL specific initializations \e must be
   done in this function: the GL context is not yet created in your viewer constructor. */
  virtual void init() { emit viewerInitialized(); };

  virtual void help();

  virtual void paintGL();

  virtual void preDraw();  
  virtual void preDrawStereo(bool leftBuffer=true);
  virtual void draw();
  virtual void fastDraw();
  virtual void postDraw();

  /*! Scene animation function. Once startAnimation() has been called, an infinite loop calls animate() and draw(). */
  virtual void animate() { emit animateNeeded(); };

  /*! Recursively call parentWidget() until the root widget is found. */
  QWidget* rootWidget() const;
  
  /*! @name OpenGL matrices */
  //@{
protected:
  /*! Simply calls Camera::loadProjectionMatrix(). Usefull for inheriting classes as it is a camera protected Camera method. */
  virtual void loadProjectionMatrix(bool reset=true) const { camera.loadProjectionMatrix(reset); };
  /*! Simply calls Camera::getProjectionMatrix(). Usefull for inheriting classes as it is a protected Camera method. */
  virtual void getProjectionMatrix(float *m) const { camera.getProjectionMatrix(m); };
  /*! Simply calls Camera::getProjectionMatrix(). Usefull for inheriting classes as it is a protected Camera method. */
  virtual void getProjectionMatrix(double *m) const { camera.getProjectionMatrix(m); };
  
  /*! Simply calls Camera::loadModelViewMatrix(). Usefull for inheriting classes as it is a camera protected Camera method. */
  virtual void loadModelViewMatrix(bool reset=true) const { camera.loadModelViewMatrix(reset); };
  /*! Simply calls Camera::getModelViewMatrix(). Usefull for inheriting classes as it is a protected Camera method. */
  virtual void getModelViewMatrix(GLdouble m[16]) const { camera.getModelViewMatrix(m); };
  /*! Simply calls Camera::getModelViewMatrix(). Usefull for inheriting classes as it is a protected Camera method. */
  virtual void getModelViewMatrix(GLfloat m[16]) const { camera.getModelViewMatrix(m); };
  //@}
  
  /*! @name Mouse and keyboard handlers */
  //@{
protected:
  virtual void mousePressEvent(QMouseEvent *);
  virtual void mouseDoubleClickEvent(QMouseEvent *);
  virtual void mouseReleaseEvent(QMouseEvent *);
  virtual void mouseMoveEvent(QMouseEvent *);
  virtual void wheelEvent(QWheelEvent *);
  virtual void timerEvent(QTimerEvent *);

  // keyboard handler
  virtual void keyPressEvent(QKeyEvent *);
  //@}
  
/*! @name Key customization */
//@{
protected:
  // Standard shortcut keys
  /*! Defines the different actions that can directly be associated with a keyboard shortcut.
   Used by setShortcutKey(), setShortcutStateKey(), shortcutKey() and shortcutStateKey().

   See the <A HREF="http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer/shortcuts.html">
  QGLViewer shortcuts page</A> for details. */
  enum ShortcutAction { DrawAxis, DrawGrid, DisplayFPS, EnableText, Exit, SaveScreenshot, \
			CameraMode, FullScreen, Stereo, Animation, Help, EditCameraPaths };
  virtual void setShortcutKey     (ShortcutAction action, int key);
  virtual void setShortcutStateKey(ShortcutAction action, ushort skey);

  Qt::Key shortcutKey(const ShortcutAction action) const;
  Qt::ButtonState shortcutStateKey(ShortcutAction action) const;

  // Key Frames shortcut keys
  virtual void setKeyFrameKey(int nb, int key);
  virtual void setAddKeyFrameStateKey(ushort skey);
  virtual void setPlayKeyFramePathStateKey(ushort skey);
  virtual void setDeleteKeyFrameStateKey(ushort skey);

  Qt::Key keyFrameKey(int nb);
  Qt::ButtonState addKeyFrameStateKey();
  Qt::ButtonState playKeyFramePathStateKey();
  Qt::ButtonState deleteKeyFrameStateKey();

  // Mouse behavior state keys
  /*! Defines the different mouse behavior modes : Camera or Frame motion.
   Use mouseStateKey() and setMouseStateKey() to select which key must be pressed to enable each one.
   See the <a href="../mouse.html">mouse page</a> for details. */
  enum MouseBehavior { CAMERA, FRAME };
  virtual void setMouseStateKey(MouseBehavior behavior, ushort skey);
  Qt::ButtonState mouseStateKey(MouseBehavior behavior) const;
//@}

  /*! @name Save and restore state  */
  //@{
public :
  virtual void saveToFile(const QString& fileName="");
  virtual void restoreFromFile(const QString& fileName="");
  
  virtual QDomElement domElement(const QString& name, QDomDocument& doc) const;
  virtual void initFromDOMElement(const QDomElement& de);
  //@}


  /*! @name Static methods for all the viewers

  All the created QGLViewers are stored in a pool. These methods allow you to use this pool. */
  //@{
  public:
  static void connectSignalToAllViewers(const QObject* sender, const char* signal, const char* member=SLOT(updateGL()));
  static void disconnectSignalToAllViewers(const QObject* sender, const char* signal, const char* member=SLOT(updateGL()));
  static void updateGLForAllViewers();
  static void saveToFileForAllViewers();
  //@}
  
private:
  // Copy constructor and opertor= are declared private and undefined
  // Prevents everyone from trying to use them
  QGLViewer(const QGLViewer& v);
  QGLViewer& operator=(const QGLViewer& v);

  // Set parameters to their default values. Called by the constructors.
  void defaultConstructor();
  // Must be call when fullScreen_ is changed
  void updateFullScreen();
  
  // D i s p l a y    f l a g s
  bool drawAxis_;	// world axis 
  bool drawGrid_;	// world XY grid 
  bool displayFPS_;	// Frame Per Seconds 
  bool enableText_;	// drawText() actually draws text or not
  bool cameraPathIsEdited_; // camera path edition mode
  bool stereo_;		// stereo display
  bool fullScreen_;	// full screen mode
  QPoint prevPos_;	// Previous window position, used for full screen mode

  // Internal debug
  bool inQGLViewerConstructorDebug_;
  // Forbids calls to updateGL() before initializeGL
  bool updateGLEnabled_;
  
  // A n i m a t i o n    p a r a m e t e r s
  bool animationStarted_; // animation mode started
  int animationPeriod_;   // period in msecs
  int animationTimerId_;
  
  // F P S    d i s p l a y 
  mutable QTime fpsTime;
  mutable unsigned int fpsCounter;
  mutable float f_p_s_;
  mutable QString fpsString_;
  
  // M a n i p u l a t e d    f r a m e 
  qglviewer::ManipulatedFrame* manipulatedFrame_;

  // D i s p l a y    l i s t s
  GLuint AxisDL_, GridDL_, CameraDL_;
  void createDisplayLists();

  // S h o r t c u t   k e y s 
  void initShortcutKeys();
  enum { nbShortcuts_ = 12 };
  Qt::Key         shortcutKey_[nbShortcuts_];
  Qt::ButtonState shortcutStateKey_[nbShortcuts_];
  // K e y   F r a m e s
  Qt::Key *keyFrameKey_;
  Qt::ButtonState addKFStateKey_, playKFStateKey_, deleteKFStateKey_;
  // M o u s e   m o d e s
  Qt::ButtonState mouseStateKey_[3];
  
  // S n a p s h o t s
  QString snapshotFilename_, snapshotFormat_;

  // Q G L V i e w e r   p o o l
  static QPtrList<QGLViewer> QGLViewerPool_;
};

#endif // QGLVIEWER_QGLVIEWER_H
