// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// QGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// QGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with QGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



// Ce programme est libre, vous pouvez le redistribuer et/ou le modifier
// selon les termes de la Licence Publique Gnrale GNU publie par la Free
// Software Foundation (version 2 ou bien toute autre version ultrieure
// choisie par vous).

// Ce programme est distribu car potentiellement utile, mais SANS AUCUNE
// GARANTIE, ni explicite ni implicite, y compris les garanties de
// commercialisation ou d'adaptation dans un but spcifique. Reportez-vous
//  la Licence Publique Gnrale GNU pour plus de dtails.

// Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en
// mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis. 

#ifndef QGLVIEWER_QUATERNION_H
#define QGLVIEWER_QUATERNION_H

#include "vec.h"
#include <math.h>
#include <iostream>

namespace qglviewer {
  /*!
   * \brief Quaternion basic class. See also the Vec documentation.
   *
   * The quaternion cooresponding to a rotation vector  v, with an angle alpha is :
   *
   *   {q[0],q[1],q[2]}  =  sin(alpha/2) * {v[0],v[1],v[2]}
   *
   *   q[3]   =  cos(alpha/2)
   */

class Quaternion
{
public :
  //! Default quaternion, a null rotation
  inline explicit Quaternion()
  { q[0]=q[1]=q[2]=0.0;  q[3]=1.0; }

  //! Simple constructor with any axis and angle (in radians)
  inline explicit Quaternion(const qglviewer::Vec& axis, const float angle)
  {
    setAxisAngle(axis, angle);
  }

  //! Constructor with normalized axis, cos & sin of half angle
  inline explicit Quaternion(const qglviewer::Vec& v, const float sin_ha, const float cos_ha)
  {
    q[0] = v[0]*sin_ha ;
    q[1] = v[1]*sin_ha ;
    q[2] = v[2]*sin_ha ;
    q[3] = cos_ha ;
  }
  
  //! Constructor from the values of a quaternion.
  inline explicit Quaternion(const float q0, const float q1, const float q2, const float q3)
  { q[0]=q0;    q[1]=q1;    q[2]=q2;    q[3]=q3; }

  //! Constructs a quaternion that will rotate from the \p from direction to the \p to direction. 
  explicit Quaternion(const qglviewer::Vec& from, const qglviewer::Vec& to);

  //! Equal operator
  inline Quaternion& operator=(const Quaternion& Q)
  {
    q[0] = Q.q[0];     q[1] = Q.q[1];    q[2] = Q.q[2];    q[3] = Q.q[3];
    return (*this);
  }

  //! Copy constructor
  inline Quaternion(const Quaternion& Q)
  { q[0] = Q.q[0];     q[1] = Q.q[1];    q[2] = Q.q[2];    q[3] = Q.q[3]; }

  QDomElement domElement(const QString& name, QDomDocument& doc) const;  
  void initFromDOMElement(const QDomElement& de);

  //! To reset a quaternion to the rotation of axis \p v and angle \p angle (in radians)
  inline void setAxisAngle(const qglviewer::Vec& v, const float angle)
  {
    const float norm = v.norm();
    if (norm < 1E-9)
      {
	// Null rotation
	q[0] = 0.0;      q[1] = 0.0;      q[2] = 0.0;      q[3] = 1.0;
      }
    else
      {
	const float sin_half_angle = sin(angle / 2.0);
	q[0] = sin_half_angle*v[0]/norm;
	q[1] = sin_half_angle*v[1]/norm;
	q[2] = sin_half_angle*v[2]/norm;
	q[3] = cos(angle / 2.0);
      }
  }

  //! Set the current quaternion value.
  inline void setValue(const float q0, const float q1, const float q2, const float q3)
  { q[0]=q0;    q[1]=q1;    q[2]=q2;    q[3]=q3; }

  void getAxisAngle(qglviewer::Vec& axis, float& angle) const;
  
  qglviewer::Vec axis() const;
  
  float angle() const;
  
  //! Bracket operator, with a constant return value
  inline float operator[](int i) const { return q[i]; }

  //! Bracket operator, returns an l-value
  inline float& operator[](int i) { return q[i]; }

  //! Returns the inverse of the Quaternion. The quaternion is \e not modified. See invert().
  inline Quaternion inverse() { return Quaternion(-q[0], -q[1], -q[2], q[3]); }

  //! Inverse the Quaternion (same rotation angle, but opposite axis). See also inverse().
  void invert() { q[0] = -q[0]; q[1] = -q[1]; q[2] = -q[2]; }

  const GLdouble* matrix() const;
  void getMatrix(GLdouble m[4][4]) const;

  void getRotationMatrix(float m[3][3]) const;
  void setFromRotationMatrix(const float m[3][3]);
  void setFromRotatedBase(const qglviewer::Vec& X,
			  const qglviewer::Vec& Y,
			  const qglviewer::Vec& Z);

  //! Returns the composition of the two rotations. Mind the order !
  inline friend const Quaternion operator*(const Quaternion& a, const Quaternion& b)
  {
    return Quaternion (a.q[3]*b.q[0] + b.q[3]*a.q[0] + a.q[1]*b.q[2] - a.q[2]*b.q[1],
		       a.q[3]*b.q[1] + b.q[3]*a.q[1] + a.q[2]*b.q[0] - a.q[0]*b.q[2],
		       a.q[3]*b.q[2] + b.q[3]*a.q[2] + a.q[0]*b.q[1] - a.q[1]*b.q[0],
		       a.q[3]*b.q[3] - b.q[0]*a.q[0] - a.q[1]*b.q[1] - a.q[2]*b.q[2]);
  }

  //! Quaternion is composed with \p q.
  inline Quaternion& operator*=(const Quaternion &q)
  {
    *this = (*this)*q;
    return *this;
  }

  //! Normalize the quaternion. Should not need to be called.
  inline float normalize()
  {
    const float norm = sqrt(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
    for (int i=0; i<4; ++i)
      q[i] /= norm;
    return norm;
  }

  //! Vec v is applied the inverse quaternion rotation (inverse of rotate()).
  inline qglviewer::Vec inverseRotate(const qglviewer::Vec& v) const
  {
    return qglviewer::Vec((1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]))*v[0] +
			  (       2.0 * (q[0] * q[1] + q[2] * q[3]))*v[1] +
			  (       2.0 * (q[2] * q[0] - q[1] * q[3]))*v[2],
			  
			  (       2.0 * (q[0] * q[1] - q[2] * q[3]))*v[0] +
			  (1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]))*v[1] +
			  (       2.0 * (q[1] * q[2] + q[0] * q[3]))*v[2],

			  (       2.0 * (q[2] * q[0] + q[1] * q[3]))*v[0] +
			  (       2.0 * (q[1] * q[2] - q[0] * q[3]))*v[1] +
			  (1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]))*v[2]  );
  }

  //! Vec \p v is applied the quaternion rotation (see also inverseRotate()).
  inline qglviewer::Vec rotate(const qglviewer::Vec& v) const
  {
    return qglviewer::Vec((1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]))*v[0] +
			  (       2.0 * (q[0] * q[1] - q[2] * q[3]))*v[1] +
			  (       2.0 * (q[2] * q[0] + q[1] * q[3]))*v[2],
    
			  (       2.0 * (q[0] * q[1] + q[2] * q[3]))*v[0] +
			  (1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]))*v[1] +
			  (       2.0 * (q[1] * q[2] - q[0] * q[3]))*v[2],

			  (       2.0 * (q[2] * q[0] - q[1] * q[3]))*v[0] +
			  (       2.0 * (q[1] * q[2] + q[0] * q[3]))*v[1] +
			  (1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]))*v[2]   );
  }
  
private: 
  //! The internal data representation is private, use []operator to acces values.
  float q[4];
};

} // namespace

std::ostream& operator<<(std::ostream& o, const qglviewer::Quaternion&);

#endif // QGLVIEWER_QUATERNION_H
