// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// QGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// QGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with QGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



// Ce programme est libre, vous pouvez le redistribuer et/ou le modifier
// selon les termes de la Licence Publique Gnrale GNU publie par la Free
// Software Foundation (version 2 ou bien toute autre version ultrieure
// choisie par vous).

// Ce programme est distribu car potentiellement utile, mais SANS AUCUNE
// GARANTIE, ni explicite ni implicite, y compris les garanties de
// commercialisation ou d'adaptation dans un but spcifique. Reportez-vous
//  la Licence Publique Gnrale GNU pour plus de dtails.

// Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en
// mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis. 

#ifndef QGLVIEWER_VEC_H
#define QGLVIEWER_VEC_H

#include <math.h>
#include <iostream>
#include <qdom.h>

// Included by all files as vec.h is at the end of the include hierarchy
#include "config.h" // Specific (win32) configuration.

namespace qglviewer {

  /*!
    \brief 3D point and vector basic class. See also the Quaternion documentation.
    
    Vec is a QGLViewer internal class. It is encapsulated in the \p qglviewer namespace.
    You can use it in your programs by specifying \p qglviewer::Vec, or by using
    the qglviewer namespace (\p using \p namespace \p qglviewer).
    
    \p Vec implements a universal explicit converter, based on the [] \p operator.
    Everywhere a \p const \p Vec& argument is expected, you can use your own 3D point type
    instead, as long as it implements the [] operator :
    \code
    float operator[] (const int i) const { return x, y or z for i=0, 1 or 2 };
    \endcode
    Indexes 0, 1 and 2 of the [] operator should respectively return the x,y and z coordinate.
    
    The following code is hence valid :
    \code
    // class myVector implements the [] operator
    myVector mv;
    fr.setPosition( qglviewer::Vec(mv) );
    // or if you use namespace qglviewer
    fr.setPosition( Vec(mv) );
    \endcode

    When Vec is used as a return value, a classical float version of the function is always available.

    You can also convert a Vec result to your own class, using code such as :
    \code
    qglviewer::Vec sc = sceneCenter();
    myVector.x() = sc.x;    // or sc[0]
    myVector.y() = sc.y;    // or sc[1]
    myVector.z() = sc.z;    // or sc[2]
    \endcode
  */
  class Vec
  {
  public :
    //! The internal data representation is public. One can use v.x, v.y, v.z.
    float x, y, z;

    //! Default constructor. Default value is (0,0,0).
    inline explicit Vec() : x(0.0), y(0.0), z(0.0) {}

    //! Standard constructor with the x,y and z values.
    inline explicit Vec(const float X, const float Y, const float Z) : x(X), y(Y), z(Z) {}

    //! Universal explicit convertor from any class to Vec. You can use your own vector class everywhere
    //! a \p const \p Vec& parameter is required, as long as it implements the [] operator.
    //! \code qglviewer::Vec pos(myVec); camera.setPosition(pos); \endcode
    template <class C>
    inline explicit Vec(const C& c) : x(c[0]), y(c[1]), z(c[2]) {}

    //! Copy constructor
    // inline explicit Vec(const Vec& v) : x(v.x), y(v.y), z(v.z) {}

    //! Operator =
    inline Vec& operator=(const Vec& v)
    {
      x = v.x;   y = v.y;   z = v.z;
      return *this;
    }

    //! Universal equal operator which allows the use of any type in place of Vec,
    //! as long as the [] operator is implemented (v[0]=v.x, v[1]=v.y, v[2]=v.z).
    // template <class C>
    // inline Vec& operator=(const C& c)
    // {
    // x=c[0]; y=c[1]; z=c[2];
    // return *this;
    // }
    
    QDomElement domElement(const QString& name, QDomDocument& doc) const;
    void initFromDOMElement(const QDomElement& de);
    
    //! Bracket operator, with a constant return value
    inline float operator[](int i) const {return (&x)[i] ;}

    //! Bracket operator, returns an l-value
    inline float& operator[](int i) {return (&x)[i] ;}

    //! The memory address of the vector. Usefull as an argument for glVectex3fv, glNormal3fv...
    inline const float* address() const { return &x; }

    //! The memory address of the vector. Usefull to directly use the address.
    inline operator const float*() const { return &x; }

    //! Set the current value without using operator equal with a temporary Vec(x,y,z).
    inline void setValue(const float X, const float Y, const float Z)
    { x=X; y=Y; z=Z; }

    //! Set the current value without using operator equal with a temporary Vec(v).
    inline void setValue(const float v)
    { x=v; y=v; z=v; }

    //! Returns the sum of the two vectors.
    inline friend Vec operator+(const Vec &a, const Vec &b)
    {
      return Vec(a.x+b.x, a.y+b.y, a.z+b.z);
    }

    //! Returns the difference of the two vectors.
    inline friend Vec operator-(const Vec &a, const Vec &b)
    {
      return Vec(a.x-b.x, a.y-b.y, a.z-b.z);
    }

    //! Unary minus operator
    inline friend Vec operator-(const Vec &a)
    {
      return Vec(-a.x, -a.y, -a.z);
    }

    //! Returns the product of the vector with a scalar.
    inline friend Vec operator*(const Vec &a, const float k)
    {
      return Vec(a.x*k, a.y*k, a.z*k);
    }

    //! Returns the product of the vector with a scalar.
    inline friend Vec operator*(float k, const Vec &a)
    {
      return Vec(a.x*k, a.y*k, a.z*k);
    }

    //! Returns the division of the vector with a scalar. With the "debug" qt \p CONFIG flag, tests for null value.
    inline friend Vec operator/(const Vec &a, const float k)
    {
#ifndef QT_NO_DEBUG
      if (fabs(k)<1.E-10)
	std::cerr << "Vec::operator / : dividing by a null value" << std::endl;
#endif
      return Vec(a.x/k, a.y/k, a.z/k);
    }

    //! Comparison based on the squared norm of the difference vector, see operator==.
    inline friend bool operator!=(const Vec &a, const Vec &b)
    {
      return !(a==b);
    }

    //! Comparison based on the \e squared norm of the difference vector, epsilon=1E-10.
    inline friend bool operator==(const Vec &a, const Vec &b)
    {
      const float epsilon = 1.0E-10;
      return (a-b).sqNorm() < epsilon;
    }

    //! Adds \p a to the vector.
    inline Vec& operator+=(const Vec &a)
    {
      x += a.x; y += a.y; z += a.z;
      return *this;
    }

    //! Substracts \p a to the vector.
    inline Vec& operator-=(const Vec &a)
    {
      x -= a.x; y -= a.y; z -= a.z;
      return *this;
    }

    //! Multiply the vector by a scalar.
    inline Vec& operator*=(float k)
    {
      x *= k; y *= k; z *= k;
      return *this;
    }

    //! Divides the vector by a scalar. With the "debug" qt \p CONFIG flag, tests for null value.
    inline Vec& operator/=(float k)
    {
#ifndef QT_NO_DEBUG
      if (fabs(k)<1.E-10)
	std::cerr << "Vec::operator /= : dividing by a null value" << std::endl;
#endif
      x /= k; y /= k; z /= k;
      return *this;
    }

    //! Dot product.
    inline friend float operator*(const Vec &a, const Vec &b)
    {
      return a.x*b.x + a.y*b.y + a.z*b.z;
    }

    //! Cross product of the two vectors. Mind the order !
    inline friend Vec cross(const Vec &a, const Vec &b)
    {
      return Vec(a.y*b.z - a.z*b.y,
		 a.z*b.x - a.x*b.z,
		 a.x*b.y - a.y*b.x);
    }

    //! Cross product of the two vectors. See also cross().  
    inline friend Vec operator^(const Vec &a, const Vec &b)
    {
      return Vec(a.y*b.z - a.z*b.y,
		 a.z*b.x - a.x*b.z,
		 a.x*b.y - a.y*b.x);
    }

    //! Returns the \e squared norm of the Vec.
    inline float sqNorm() const { return x*x + y*y + z*z; }

    //! Returns the norm of the vector.
    inline float norm() const { return sqrt(x*x + y*y + z*z); }

    //! Normalizes the Vec and returns the norm value.
    inline Vec& normalize()
    {
      const float n = norm();
#ifndef QT_NO_DEBUG
      if (n<1.E-10)
	std::cerr << "Vec::normalize : normalizing a null vector" << std::endl;
#endif
      *this /= n;
      return *this;
    }
    
    void projectOnAxis (const Vec& dir);
    void projectOnPlane(const Vec& n);
  };
  
} // namespace

std::ostream& operator<<(std::ostream& o, const qglviewer::Vec&);
  
#endif // QGLVIEWER_VEC_H
